package adt.scanner;

import adt.entity.Host;
import java.util.*;
import adt.service.PortScannerService;
import adt.service.WrittingService;
import com.google.common.net.InetAddresses;
import java.net.NetworkInterface;
import com.stealthcopter.networktools.ARPInfo;
import lombok.extern.slf4j.Slf4j;
import lombok.Getter;
import java.util.concurrent.CountDownLatch;
import java.util.Stack;
import java.io.IOException;
import java.io.*;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.tomcat.jni.File;
import java.net.SocketException;
import java.util.function.Consumer;
import java.util.stream.Stream;

@Slf4j
@Getter
public class HostScanner {

    private String subnetIPAddress;
    private Integer subnetMask;
    private int timeout;
    private Consumer<Host> onHostFound;
    private final PortScannerService portScannerService;
    List<Host> hostsList;

    private String netAddress;
    private String broadAddress;

    /**
     * Performs host scan with {@code this} configuration
     * @return
     */
    public void scan() {

        getNetworkAndBroadcast(subnetIPAddress, subnetMask);

        CountDownLatch latch = new CountDownLatch((int) Math.pow(2, 32 - subnetMask) - 2);
        ThreadPoolExecutor executor = (ThreadPoolExecutor) Executors.newFixedThreadPool((int) Math.pow(2, 32 - subnetMask) - 1);

        generateIPs().forEach(addr -> executor.submit(() -> {

            // PUEDE QUE LO USEMOS: ip -o -f inet addr show | awk '/scope global/ {print $4}' > test.txt
            if (addrIsReachable(addr, timeout) && !addr.getHostAddress().equals(netAddress) && !addr.getHostAddress().equals(broadAddress) ) {


                    // TimestampINI device
                    Date iniTimestamp_device = new Date();

                    String mac = ARPInfo.getMACFromIPAddress(addr.getHostAddress());
                    log.info("MAC for this ip ({}): {}", addr.getHostAddress(), mac);
                    if (mac == null) {

                        try {
                            mac = checkNetworkInterfaces(addr.getHostAddress());
                        } catch (Exception e) {

                        }
                        //log.info("No tenia pero ahora MAC for this ip ({}): {}", addr.getHostAddress(), mac);
                    }


                    Host host = Host.createHost(addr.getHostAddress(), addr.getCanonicalHostName(), mac, "scan");
                    host.setTimestamp_ini(iniTimestamp_device);

                    String macManufacturer = MacScanner.getMacScanner().getMacManufacturer(host);
                    host.setMacManufacturer(macManufacturer);
                    log.info("Host {} found with mac {} manufactured by {}", host.getIp(), host.getMac(), host.getMacManufacturer());

                    // Once found, scan this host's ports
                    portScannerService.scanHost(host, "scan");

                    synchronized (hostsList) {
                        hostsList.add(host);
                    }

                }

            latch.countDown();

        }));

        try {
            latch.await();
        } catch (InterruptedException E) {
            // handle
        }

        executor.shutdown();
    }

    /**
     *
     * @param addr The address to be checked
     * @param timeout Timeout in miliseconds
     * @return Whether {@code addr} is reachable or not
     */
    private static boolean addrIsReachable(InetAddress addr, int timeout) {

        if(addr != null){
            try {
                return addr.isReachable(timeout);
            } catch (IOException e) {
                //e.printStackTrace();
                //log.warn("SOCKET EXCEPTION HANDLED");
            }
        }
        return false;
    }

    /**
     * Callback method executed every time a new host is found.
     *
     * @param onHostFound Void function receiving a {@link adt.entity.Host Host} as a parameter
     * @return {@code this} for chaining
     */
    public HostScanner setOnHostFound(Consumer<Host> onHostFound) {
        this.onHostFound = onHostFound;
        return this;
    }

    /**
     * Set the timeout for {@code this} scanner.
     *
     * @param timeout Timeout in miliseconds
     * @return {@code this} for chaining
     */
    public HostScanner setTimeout(int timeout) {
        if (timeout < 1) throw new IllegalArgumentException("Timeout must be an integer greater than 0.");
        this.timeout = timeout;
        return this;
    }

    /**
     * Generate all the possible IPs for {@code this} configuration. The stream generated will be limited to the maximum
     * number of devices on the subnet specified to {@code this}. Also, the first address is skipped as it should be a
     * network address.
     *
     * @return A spliterator for the generated stream
     */
    private Stream<InetAddress> generateIPs() {
        long limit = (long) (Math.pow(2, 32 - subnetMask) - 1);
        InetAddress firstAddress = InetAddresses.forString(subnetIPAddress);

        return Stream.iterate(firstAddress, InetAddresses::increment)
                .limit(limit);
    }

    /**
     * @param subnetAddress A subnet address with or without mask specified from where the scan will start
     * @return A new {@link HostScanner}
     */
    public static HostScanner fromSubnetAddress(String subnetAddress, PortScannerService portScannerService) {

        String subnetIPAddress = subnetAddress;
        int subnetMask = 24;

        if (subnetAddress.contains("/")) {
            String[] split = subnetAddress.split("/");
            subnetIPAddress = split[0];
            subnetMask = Integer.parseInt(split[1]);
        }

        if (!InetAddresses.isInetAddress(subnetIPAddress))
            throw new IllegalArgumentException(subnetIPAddress + " does not seem to be a valid IP address");

        if (subnetMask < 0 || subnetMask > 32)
            throw new IllegalArgumentException("/" + subnetMask + " is not a valid mask");

        return new HostScanner(subnetIPAddress, subnetMask, portScannerService);

    }

    /**
     * Creates a new {@link HostScanner} instance.
     * <br><br>
     * {@code timeout} default value is 1000
     * <br>
     * {@code onHostFound} default value is {@code System.out::println}
     *
     * @param subnetIPAddress Subnet IP Address
     * @param subnetMask      Subnet mask
     */
    private HostScanner(String subnetIPAddress, Integer subnetMask, PortScannerService portScannerService) {
        this.subnetIPAddress = subnetIPAddress;
        this.subnetMask = subnetMask;
        this.timeout = 1000;
        this.onHostFound = System.out::println;
        this.portScannerService = portScannerService;
        this.hostsList = new ArrayList<>();
    }

    private int[] bina(String[] str)
    {
        int re[] = new int[32];
        int a, b, c, d, i, rem;
        a = b = c = d = 1;
        Stack<Integer> st = new Stack<Integer>();

        // Separate each number of the IP address
        if (str != null)
        {
            a = Integer.parseInt(str[0]);
            b = Integer.parseInt(str[1]);
            c = Integer.parseInt(str[2]);
            d = Integer.parseInt(str[3]);
        }

        // convert first number to binary
        for (i = 0; i <= 7; i++)
        {
            rem = a % 2;
            st.push(rem);
            a = a / 2;
        }

        // Obtain First octet
        for (i = 0; i <= 7; i++) {
            re[i] = st.pop();
        }

        // convert second number to binary
        for (i = 8; i <= 15; i++) {
            rem = b % 2;
            st.push(rem);
            b = b / 2;
        }

        // Obtain Second octet
        for (i = 8; i <= 15; i++) {
            re[i] = st.pop();
        }

        // convert Third number to binary
        for (i = 16; i <= 23; i++) {
            rem = c % 2;
            st.push(rem);
            c = c / 2;
        }

        // Obtain Third octet
        for (i = 16; i <= 23; i++) {
            re[i] = st.pop();
        }

        // convert fourth number to binary
        for (i = 24; i <= 31; i++) {
            rem = d % 2;
            st.push(rem);
            d = d / 2;
        }

        // Obtain Fourth octet
        for (i = 24; i <= 31; i++) {
            re[i] = st.pop();
        }

        return (re);
    }

    // cls returns class of given IP address
    private char cls(String[] str)
    {
        int a = Integer.parseInt(str[0]);
        if (a >= 0 && a <= 127)
            return ('A');
        else if (a >= 128 && a <= 191)
            return ('B');
        else if (a >= 192 && a <= 223)
            return ('C');
        else if (a >= 224 && a <= 239)
            return ('D');
        else
            return ('E');
    }

    // Converts IP address
    // from binary to decimal form
    private int[] deci(int[] bi)
    {

        int[] arr = new int[4];
        int a, b, c, d, i, j;
        a = b = c = d = 0;
        j = 7;

        for (i = 0; i < 8; i++) {

            a = a + (int)(Math.pow(2, j)) * bi[i];
            j--;
        }

        j = 7;
        for (i = 8; i < 16; i++) {

            b = b + bi[i] * (int)(Math.pow(2, j));
            j--;
        }

        j = 7;
        for (i = 16; i < 24; i++) {

            c = c + bi[i] * (int)(Math.pow(2, j));
            j--;
        }

        j = 7;
        for (i = 24; i < 32; i++) {

            d = d + bi[i] * (int)(Math.pow(2, j));
            j--;
        }

        arr[0] = a;
        arr[1] = b;
        arr[2] = c;
        arr[3] = d;
        return arr;
    }

    private void getNetworkAndBroadcast(String ip, int mask){

        int i;
        String[] str = new String[4];

        String tr = ip;//str1[0];

        // Split IP address into 4 subparts x, y, z, t
        str = tr.split("\\.");

        int[] b = new int[32];

        // Convert IP address to binary form
        b = bina(str);

        int n = mask;//Integer.parseInt(mask);
        int[] ntwk = new int[32];
        int[] brd = new int[32];
        int t = 32 - n;

        // Obtanining network address
        for (i = 0; i <= (31 - t); i++) {

            ntwk[i] = b[i];
            brd[i] = b[i];
        }

        // Set 32-n bits to 0
        for (i = 31; i > (31 - t); i--) {

            ntwk[i] = 0;
        }

        // Obtaining Broadcast address
        // by setting 32-n bits to 1
        for (i = 31; i > (31 - t); i--) {

            brd[i] = 1;
        }

        // Converting network address to decimal
        int[] nt = deci(ntwk);

        // Converting broadcast address to decimal
        int[] br = deci(brd);


        netAddress = nt[0] + "." + nt[1] + "." + nt[2] + "." + nt[3];

        broadAddress = br[0] + "." + br[1] + "." + br[2] + "." + br[3];
    }
    

    // UNUSED
    private void checkARPTable(){

        String[] arpCheck = {"arp", "-a"};
        try {
            Process p = new ProcessBuilder(arpCheck).start();
            BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
            String arpOut = null;

            while ((arpOut = in.readLine()) != null) {

               log.info("FOUND MAC {}", arpOut);
            }
        }catch(Exception io){
            //handled
        }
    }

    /**************/
    public String checkNetworkInterfaces(String toMatch) throws SocketException {
        Enumeration<NetworkInterface> nets = NetworkInterface.getNetworkInterfaces();
        String mac = null;
        for (NetworkInterface netint : Collections.list(nets)) {
            mac = displayInterfaceInformation(netint, toMatch);
            if (mac.length() > 0 ) {
                break;
                //return netName;
            }
        }

        return mac;
    }

    /**************/
    public String displayInterfaceInformation(NetworkInterface netint, String toMatch) throws SocketException {

        Enumeration<InetAddress> inetAddresses = netint.getInetAddresses();
        String matched = null;
        String macToStore = "";
        for (InetAddress inetAddress : Collections.list(inetAddresses)) {

            if(inetAddress.getHostAddress().equals(toMatch)){

                matched = netint.getName();

                byte[] mac = netint.getHardwareAddress();


                if(mac != null) {

                    for (int i = 0; i < mac.length; i++) {

                        String toCheck = Integer.toHexString(mac[i]);
                        if(toCheck.length() > 2){

                            toCheck = toCheck.substring(toCheck.length()-3, toCheck.length()-1);
                        }

                        if(i < mac.length-1){
                            macToStore = macToStore + toCheck + ":";
                        }else{
                            macToStore = macToStore + toCheck;
                        }
                    }

                }

                break;
            }
        }

        return macToStore;
    }



}


